use std::collections::HashMap;

use ed25519_dalek as dalek;
use ed25519_dalek::Verifier;
use rand::prelude::random;
use rs_merkle::{MerkleProof, MerkleTree};
use sha2::{Digest, Sha256};

pub use rs_merkle::algorithms::Sha256 as DefaultHasher; // Imports from sha2 crate

/// TODO: Use hashes instead of public addresses.
/// TODO: Don't forget document endiannes.
/// TODO: Allow doubly signed obligations or implement abstract interfaces for Cert (see below).

/// Implement abstract Cert datatype to allow attestation of hyper-relations:
/// TODO: Use serializable trait to generalize over the subject field.
/// TODO: Allow per subject type verification trait.  Hard not to go
/// full crypto-conditions but better not to complicate things too
/// much.

pub type Address = [u8; 32];

#[derive(Copy, Clone, PartialEq, Debug)]
pub struct Obligation {
    pub from: PseudoAnonID,
    pub to: PseudoAnonID,
    pub value: u32,
}

impl Obligation {
    fn serialize(self) -> Vec<u8> {
        [&self.from[..], &self.value.to_le_bytes()[..], &self.to[..]].concat()
    }
}

pub type Signature = [u8; 64];
pub type Salt = [u8; 32];

#[derive(Copy, Clone)]
pub struct AuthCert {
    pub subject: Obligation,
    pub signature: Signature,
}

impl AuthCert {
    /// TODO: add optional salt argument
    pub fn new(sub: Obligation, sig: Signature) -> Result<Self, &'static str> {
        match dalek::VerifyingKey::from_bytes(&sub.from)
            .unwrap()
            .verify(&sub.serialize()[..], &dalek::Signature::from_bytes(&sig))
        {
            Ok(_) => Ok(Self {
                subject: sub,
                signature: sig,
            }),
            Err(_) => Err("Invalid signature"),
        }
    }
}

pub type PseudoAnonID = [u8; 32];

fn pseudo_anon_id(a: Address) -> PseudoAnonID {
    let mut h = Sha256::new();
    h.update(a);
    let res: PseudoAnonID = h.finalize().into();
    res
}

pub struct Members {
    pub data: MerkleTree<DefaultHasher>,
    indicies: HashMap<PseudoAnonID, usize>,
}

pub struct Member {
    pub id: PseudoAnonID,
    index: usize,
}

pub struct MembershipProof {
    subjects: Vec<Member>,
    root: [u8; 32],
    hashes: MerkleProof<DefaultHasher>,
}

impl MembershipProof {
    pub fn verify(&self) -> bool {
        self.hashes.verify(
            self.root,
            &self.subjects.iter().map(|x| x.index).collect::<Vec<_>>(),
            &self.subjects.iter().map(|x| x.id).collect::<Vec<_>>(),
            self.subjects.len(),
        )
    }
}

impl Members {
    pub fn add(mut self, m: Address) {
        self.data.insert(m).commit()
    }
    pub fn proof(&self, subjects: Vec<PseudoAnonID>) -> Result<MembershipProof, &'static str> {
        if (1..subjects.len()).any(|i| subjects[i..].contains(&subjects[i - 1])) {
            return Err("Duplicate subjects");
        }
        let (subjects, indices) = {
            let mut tosort = subjects
                .iter()
                .map(|x| Member {
                    id: *x,
                    index: *self.indicies.get(x).unwrap(),
                })
                .collect::<Vec<_>>();
            tosort.sort_by_key(|x| x.index);
            let indices: Vec<usize> = Vec::from_iter(tosort.iter().map(|x| x.index));
            (tosort, indices)
        };
        Ok(MembershipProof {
            subjects: subjects,
            hashes: self.data.proof(&indices),
            root: self.data.root().unwrap(),
        })
    }
    pub fn index(&self, member: PseudoAnonID) -> Option<usize> {
        match self.indicies.get(&member) {
            Some(i) => Some(*i),
            None => None,
        }
    }
}

pub struct Cert {
    pub subject: Obligation,
    pub authenticity: AuthCert,
    pub membership: MembershipProof,
    pub salt: Salt,
    pub hash: Vec<u8>,
}

impl Cert {
    pub fn new(sub: Obligation, auth: AuthCert, mem: MembershipProof) -> Self {
        let mut h = Sha256::new();
        let salt: Salt = random();
        h.update(sub.serialize());
        h.update(salt);
        Self {
            subject: sub,
            authenticity: auth,
            membership: mem,
            salt: salt,
            hash: h.finalize().to_vec(),
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use ed25519_dalek::Signer;
    use once_cell::sync::Lazy;
    use rand::seq::IteratorRandom;
    use std::collections::HashSet;

    #[allow(dead_code)]
    pub fn make_address() -> Address {
        random()
    }

    fn sk_gen() -> (dalek::SigningKey, [u8; dalek::SECRET_KEY_LENGTH], Address) {
        let skobj = dalek::SigningKey::from_bytes(&random());
        (
            skobj.clone(),
            skobj.to_bytes(),
            skobj.verifying_key().to_bytes(),
        )
    }

    #[allow(dead_code)]
    struct ObligationFixture {
        o: Obligation,
        f: dalek::SigningKey,
        t: dalek::SigningKey,
        fpk: [u8; 32],
        tpk: [u8; 32],
        fsk: [u8; 32],
        tsk: [u8; 32],
    }

    static OFIX: Lazy<ObligationFixture> = Lazy::new(|| {
        let (f, fsk, fpk) = sk_gen();
        let (t, tsk, tpk) = sk_gen();
        ObligationFixture {
            o: Obligation {
                from: fpk,
                to: tpk,
                value: 1000,
            },
            f: f,
            t: t,
            fpk: fpk,
            tpk: tpk,
            fsk: fsk,
            tsk: tsk,
        }
    });

    #[allow(dead_code)]
    struct AddrMeta {
        sk: dalek::SigningKey,
        id: PseudoAnonID,
    }

    static ADDRESSES: Lazy<HashMap<Address, AddrMeta>> = Lazy::new(|| {
        let mut v: HashMap<Address, AddrMeta> = HashMap::new();
        for _i in 0..10 {
            let (obj, _, pk) = sk_gen();
            v.insert(
                pk,
                AddrMeta {
                    sk: obj,
                    id: pseudo_anon_id(pk),
                },
            );
        }
        v
    });

    static MEMBERS: Lazy<Members> = Lazy::new(|| {
        let tree = MerkleTree::<DefaultHasher>::from_leaves(
            &ADDRESSES.keys().cloned().collect::<Vec<PseudoAnonID>>(),
        );
        let mut indices: HashMap<PseudoAnonID, usize> = HashMap::with_capacity(tree.leaves_len());
        tree.leaves()
            .unwrap()
            .iter()
            .enumerate()
            .for_each(|(i, x)| {
                indices.insert(*x, i);
            });
        Members {
            data: tree,
            indicies: indices,
        }
    });

    fn rand_member() -> Option<PseudoAnonID> {
        MEMBERS
            .indicies
            .keys()
            .choose(&mut rand::thread_rng())
            .cloned()
    }

    fn rand_unique_member_gen() -> Box<impl FnMut() -> PseudoAnonID> {
        let mut cache: HashSet<PseudoAnonID> = HashSet::new();
        // note: will go in endless loop if the cache >= set of members
        Box::new(move || loop {
            let rmem = rand_member();
            if cache.get(&rmem.unwrap()).is_some() {
                continue;
            }
            cache.insert(rmem.unwrap());
            break rmem.unwrap();
        })
    }

    fn sort_by_indices(m: Vec<PseudoAnonID>) -> Vec<PseudoAnonID> {
        let mut res = m.clone();
        res.sort_by_cached_key(|x| MEMBERS.indicies.get(x));
        res
    }

    fn rand_members(n: usize) -> Vec<PseudoAnonID> {
        let l = if n > ADDRESSES.len() {
            ADDRESSES.len()
        } else {
            n
        };
        let mut f = rand_unique_member_gen();
        sort_by_indices(std::iter::from_fn(|| Some(f())).take(l).collect())
    }

    fn rand_obligation() -> Obligation {
        let mut r = rand_unique_member_gen();
        Obligation {
            from: r(),
            to: r(),
            value: random(),
        }
    }

    #[test]
    fn serialize_obligation() {
        assert_eq!(
            Obligation {
                from: [1; 32],
                to: [2; 32],
                value: 4_294_967_295, // Max u32 (255 255 255 255)
            }
            .serialize(),
            [
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                1, 1, 1, 1, 255, 255, 255, 255, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2
            ]
        );
    }

    #[test]
    fn authenticity_cert_new() {
        let sig = OFIX.f.sign(&OFIX.o.serialize()[..]).to_bytes();
        let cert = AuthCert::new(OFIX.o, sig);
        assert!(cert.is_ok());
        assert_eq!(cert.unwrap().subject, OFIX.o);
        assert_eq!(cert.unwrap().signature, sig);
        assert!(AuthCert::new(OFIX.o, [0; 64]).is_err());
    }

    #[test]
    fn membership_cert_new() {
        let o = rand_obligation();
        let _c = Cert::new(
            o,
            AuthCert::new(
                o,
                ADDRESSES
                    .get(&o.from)
                    .unwrap()
                    .sk
                    .sign(&o.serialize())
                    .into(),
            )
            .expect("Auth failed"),
            MEMBERS.proof([o.from, o.to].to_vec()).unwrap(),
        );
    }

    // FIXME: Failing test - some weirdnes in rs_merkle or indexes fiddling.
    #[test]
    #[ignore]
    fn test_merkle() {
        assert!(MEMBERS.proof(rand_members(2)).unwrap().verify());
    }
}
